//! An extension that bridges X11 with OpenGL, an API used for rendering
//! graphics with the GPU.

mod context;
mod drawable;

use x11::glx;

use crate::errors::Result;
use crate::xlib::{Display, VisualInfo};

pub use self::context::GlxContext;
pub use self::drawable::GlxDrawable;

pub trait GlxExtension {
  /// Checks if glx is available in this implementation.
  fn query_glx_extension(&self) -> Result<bool>;

  fn create_context(
    &self,
    visual_info: &VisualInfo,
    share_list: Option<&GlxContext>,
    direct: bool,
  ) -> Result<GlxContext>;

  fn make_current(&self, drawable: GlxDrawable, ctx: GlxContext) -> Result<()>;

  fn choose_visual(&self, screen: i32) -> Result<VisualInfo>;
}

impl GlxExtension for Display {
  fn query_glx_extension(&self) -> Result<bool> {
    let result =
      unsafe { glx::glXQueryExtension(self.inner, 0 as *mut _, 0 as *mut _) };
    Ok(result == 1)
  }

  fn create_context(
    &self,
    visual_info: &VisualInfo,
    share_list: Option<&GlxContext>,
    direct: bool,
  ) -> Result<GlxContext> {
    let share_list = share_list
      .map(|ctx| ctx.as_raw())
      .unwrap_or_else(|| 0 as *mut _);
    let ctx = unsafe {
      glx::glXCreateContext(
        self.as_raw(),
        visual_info.as_raw(),
        share_list,
        direct.into(),
      )
    };
    Ok(GlxContext { display: self, ctx })
  }

  fn make_current(&self, drawable: GlxDrawable, ctx: GlxContext) -> Result<()> {
    unsafe {
      glx::glXMakeCurrent(self.as_raw(), drawable.as_raw(), ctx.as_raw())
    };
    Ok(())
  }

  fn choose_visual(&self, screen: i32) -> Result<VisualInfo> {
    let visual =
      unsafe { glx::glXChooseVisual(self.as_raw(), screen, 0 as *mut _) };
    Ok(VisualInfo::from_raw(visual))
  }
}
