use std::ffi::CString;
use std::mem::MaybeUninit;
use std::os::raw::c_char;
use std::{ptr, env};

use x11::xlib;

use crate::errors::{Error, Result};
use crate::ffi;

use super::cursor::Cursor;
use super::event::Event;
use super::screen::Screen;
use super::visual::Visual;
use super::window::Window;

/// A connection to an X server.
pub struct Display {
  pub(crate) inner: *mut xlib::Display,
}

pub struct Grab(pub(crate) *mut xlib::Display);

/// Something that's part of a display.
pub trait GetDisplay {
  /// Get the current display
  fn get_display(&self) -> *mut xlib::Display;
}

impl Display {
  /// Opens a new connection to an X server.
  ///
  /// On POSIX-conformant systems, the display name or DISPLAY environment
  /// variable can be a string in the format:
  ///
  ///     hostname:number.screen_number
  pub fn connect(display_name: impl AsRef<str>) -> Result<Display> {
    let display_name = CString::new(display_name.as_ref()).unwrap();
    let inner = unsafe { xlib::XOpenDisplay(display_name.as_ptr()) };
    if inner.is_null() {
      return Err(Error::DisplayOpenError);
    }
    Ok(Display { inner })
  }

  pub fn connect_default() -> Result<Display> {
    let env = env::var("DISPLAY")?;
    Self::connect(env)
  }

  /// Create a Display for an existing connection
  pub fn from_handle(handle: u64) -> Self {
    Display {
      inner: handle as *mut xlib::Display,
    }
  }

  /// Grab
  pub fn grab(&self) -> Grab {
    unsafe { xlib::XGrabServer(self.inner) };
    Grab(self.inner)
  }

  /// Wrapper around XCreateFontCursor
  pub fn create_font_cursor(&self, shape: u32) -> Result<Cursor> {
    let cursor =
      unsafe { xlib::XCreateFontCursor(self.inner, shape) as xlib::Cursor };
    if cursor == 0 {
      return Err(Error::CreateCursorError);
    }
    Ok(Cursor {
      display: self,
      inner: cursor,
    })
  }

  /// Get the next event, blocks until an event is reached.
  pub fn next_event(&self) -> Result<Event> {
    debug!("fishing for next event...");
    let mut event = MaybeUninit::uninit();
    let ptr = event.as_mut_ptr();
    unsafe { xlib::XNextEvent(self.inner, ptr) };
    // TODO: check to make sure this isn't null
    let event = unsafe { event.assume_init() };
    debug!("event: {:?}", event);
    unsafe { Event::from_raw(event) }
  }

  /// Returns the number of events that are still pending
  pub fn pending(&self) -> Result<i32> {
    Ok(unsafe { xlib::XPending(self.inner) })
  }

  /// Gets the raw X Display handle
  pub fn as_raw(&self) -> *mut xlib::Display {
    self.inner
  }

  /// Gets the default visual
  pub fn default_visual(&self, screen: &Screen) -> Visual {
    let visual = unsafe { xlib::XDefaultVisual(self.inner, screen.0) };
    Visual { inner: visual }
  }

  /// Get the default screen using `XDefaultScreen`
  pub fn get_default_screen(&self) -> Result<Screen> {
    let screen = unsafe { xlib::XDefaultScreen(self.inner) };
    Ok(Screen(screen))
  }

  /// Returns the root window for the given screen.
  pub fn get_root_window(&self, screen: &Screen) -> Result<Window> {
    let inner = unsafe { xlib::XRootWindow(self.inner, screen.0) };
    if inner == 0 {
      return Err(Error::GetWindowError);
    }
    let window = Window {
      display: self,
      inner,
    };
    Ok(window)
  }

  /// Returns the root window for the default screen.
  pub fn get_default_root_window(&self) -> Result<Window> {
    let inner = unsafe { xlib::XDefaultRootWindow(self.inner) };
    if inner == 0 {
      return Err(Error::GetWindowError);
    }
    let window = Window {
      display: self,
      inner,
    };
    Ok(window)
  }

  /// Translate coordinates relative to w1 to coordinates relative to w2.
  /// If the coordinates are contained in a mapped child of the destination
  /// window, the third return value will hold that child window.
  pub fn translate_coordinates(
    &self,
    w1: Window,
    x: i32,
    y: i32,
    w2: Window,
  ) -> Result<(i32, i32, Option<Window>)> {
    let mut rx = 0;
    let mut ry = 0;
    let mut child_return: xlib::Window = 0;
    let status = unsafe {
      xlib::XTranslateCoordinates(
        self.inner,
        w1.inner,
        w2.inner,
        x,
        y,
        &mut rx,
        &mut ry,
        &mut child_return,
      )
    };
    if status == 0 {
      return Err(Error::TranslateCoordinatesError);
    }
    let child = match child_return {
      0 => None,
      val => Some(Window {
        display: self,
        inner: val,
      }),
    };
    Ok((rx, ry, child))
  }

  /// Sync
  pub fn sync(&self, discard: bool) {
    unsafe { xlib::XSync(self.inner, if discard { 1 } else { 0 }) };
  }

  /// Returns the focus window and the current focus state.
  pub fn get_input_focus(&self) -> Result<(Window, i32)> {
    let mut focus_return: xlib::Window = 0;
    let mut revert_to_return = 0;
    unsafe {
      xlib::XGetInputFocus(self.inner, &mut focus_return, &mut revert_to_return)
    };
    let window = Window {
      display: self,
      inner: focus_return,
    };
    return Ok((window, revert_to_return));
  }

  /// Query extension
  pub fn query_extension(&self, name: impl AsRef<str>) -> Result<()> {
    let name = ffi::c_char_star_from_string(name)?;
    let major_opcode_return = ptr::null_mut();
    let first_event_return = ptr::null_mut();
    let first_error_return = ptr::null_mut();
    let _result = unsafe {
      xlib::XQueryExtension(
        self.inner,
        name,
        major_opcode_return,
        first_event_return,
        first_error_return,
      )
    };
    // TODO: check reuslt

    unimplemented!()
  }

  /// List extensions
  pub fn list_extensions(&self) -> Result<ListExtensions> {
    let nextensions_return = ptr::null_mut();
    let result =
      unsafe { xlib::XListExtensions(self.inner, nextensions_return) };
    // TODO: check result null
    // TODO: check nextensions_return
    let nextensions_return = unsafe { *nextensions_return } as isize;
    Ok(ListExtensions(&self, result, nextensions_return))
  }
}

impl Drop for Display {
  fn drop(&mut self) {
    unsafe { xlib::XCloseDisplay(self.inner) };
  }
}

impl Drop for Grab {
  fn drop(&mut self) {
    unsafe { xlib::XUngrabServer(self.0) };
  }
}

pub struct ListExtensions<'a>(&'a Display, *mut *mut c_char, isize);

impl<'a> ListExtensions<'a> {
  pub fn iter(&self) -> ListExtensionsIter {
    ListExtensionsIter(self, 0)
  }
}

pub struct ListExtensionsIter<'a>(&'a ListExtensions<'a>, isize);

impl<'a> Iterator for ListExtensionsIter<'a> {
  type Item = String;
  fn next(&mut self) -> Option<Self::Item> {
    if self.1 >= (self.0).2 {
      return None;
    }

    // TODO: check for null
    let base = unsafe { *(self.0).1.offset(self.1) };
    let s = unsafe { ffi::string_from_c_char_star(base).unwrap() };
    // TODO: don't unwrap here
    Some(s)
  }
}
