use x11::xlib;

use crate::errors::{Error, Result};

/// A handle to an XImage
///
/// Xlib provides some structures to perform operations on images. However,
/// because there are different in-memory representations, working with images
/// is actually totally fucked. If you want to maintain your sanity, do NOT pry
/// into the raw data buffer of an Image. Instead, index into the image using
/// `PixBuffer::get_pixel` which already performs the conversion and returns a
/// tuple of RGB values.
///
/// Peeking into the raw data buffer of an Image
/// --------------------------------------------
///
/// Images consist of lines of longs in row-major form.
pub struct Image {
  pub inner: *mut xlib::XImage,
}

/// Image byte order
pub enum ImageByteOrder {
  /// Least significant byte first
  LSBFirst,
  /// Most significant byte first
  MSBFirst,
}

impl Image {
  /// Get the size (in bytes) of the data buffer.
  #[inline]
  pub fn get_size(&self) -> usize {
    4 * self.get_width() as usize * self.get_height() as usize
  }

  /// Get the image width
  #[inline]
  pub fn get_width(&self) -> u32 {
    unsafe { (*self.inner).width as u32 }
  }

  /// Get the image height
  #[inline]
  pub fn get_height(&self) -> u32 {
    unsafe { (*self.inner).height as u32 }
  }

  /// Get the image depth
  #[inline]
  pub fn get_depth(&self) -> u32 {
    unsafe { (*self.inner).depth as u32 }
  }

  /// Get byte order (LSB or MSB)
  #[inline]
  pub fn get_byte_order(&self) -> Result<ImageByteOrder> {
    let byte_order = unsafe { (*self.inner).byte_order };
    match byte_order {
      xlib::LSBFirst => Ok(ImageByteOrder::LSBFirst),
      xlib::MSBFirst => Ok(ImageByteOrder::MSBFirst),
      order => Err(Error::InvalidByteOrder(order)),
    }
  }

  fn get_rgb_masks(&self) -> (u64, u64, u64) {
    let red_mask = unsafe { (*self.inner).red_mask };
    let green_mask = unsafe { (*self.inner).green_mask };
    let blue_mask = unsafe { (*self.inner).blue_mask };
    (red_mask, blue_mask, green_mask)
  }

  /// Produces a PixBuffer.
  pub fn buffer(&self) -> PixBuffer {
    let size = self.get_size();
    let buf = unsafe { (*self.inner).data as *mut u8 };
    let masks = self.get_rgb_masks();
    PixBuffer {
      buf,
      size,
      masks,
      image: self,
    }
  }
}

impl Drop for Image {
  fn drop(&mut self) {
    unsafe { xlib::XDestroyImage(self.inner) };
  }
}

/// The buffer pointed to by an XImage.
pub struct PixBuffer<'a> {
  /// The raw pointer to the buffer
  pub buf: *mut u8,

  /// The size of the buffer
  pub size: usize,

  /// Rgb masks
  pub masks: (u64, u64, u64),

  pub image: &'a Image,
}

impl<'a> PixBuffer<'a> {
  /// Retrieve a direct reference to the internal buffer.
  pub fn get_buffer(&self) -> &[u8] {
    unsafe { std::slice::from_raw_parts(self.buf, self.size) }
  }

  /// Gets the rgb value of a particular pixel, after performing conversion
  ///
  /// This uses XImage's get_pixel function to retrieve the pixel value, and
  /// then the various red/green/blue masks to actually perform the
  /// conversions, and fails (returns `None`) if the (x, y) coordinates given
  /// aren't actually in the image.
  #[inline]
  pub fn get_pixel(&self, x: u32, y: u32) -> Option<(u8, u8, u8)> {
    let get_pixel = unsafe { (*self.image.inner).funcs.get_pixel }.unwrap();
    let pixel = unsafe { get_pixel(self.image.inner, x as i32, y as i32) };

    let red_mask = unsafe { (*self.image.inner).red_mask };
    let green_mask = unsafe { (*self.image.inner).green_mask };
    let blue_mask = unsafe { (*self.image.inner).blue_mask };

    let red = ((pixel & red_mask) >> 16) as u8;
    let green = ((pixel & green_mask) >> 8) as u8;
    let blue = ((pixel & blue_mask) >> 0) as u8;

    Some((red, green, blue))
  }
}
