use std::mem;
use std::os::raw::{c_long, c_ulong};

use x11::xlib;

use crate::errors::{Error, Result};
use crate::rect::Rectangle;
use crate::xlib::{
  Atom, Colormap, Cursor, Display, Drawable, GetDisplay, Image, Inherit,
  IntoInt, Noneable, Pixmap,
};

/// A wrapper around a window handle.
#[derive(Clone)]
pub struct Window<'a> {
  pub display: &'a Display,
  pub inner: xlib::Window,
}

impl<'a> Window<'a> {
  /// Create a new window
  pub fn create(
    display: &'a Display,
    parent: Option<Window>,
    location: Rectangle,
    set_attributes: SetWindowAttributes,
  ) -> Result<Window<'a>> {
    let parent = match parent {
      Some(parent) => parent,
      None => display.get_default_root_window()?,
    };
    let screen = display.get_default_screen()?;
    let visual = display.default_visual(&screen);
    let window = unsafe {
      xlib::XCreateWindow(
        display.as_raw(),
        parent.as_raw(),
        location.x as i32,
        location.y as i32,
        location.width,
        location.height,
        0,
        0,
        0,
        visual.as_raw(),
        0,
        set_attributes.into_raw(),
      )
    };
    Ok(Window {
      display,
      inner: window,
    })
  }

  /// Create a new Window instance from an existing ID
  pub fn create_from_handle(display: &Display, id: u64) -> Result<Window> {
    Ok(Window { display, inner: id })
  }

  /// Get window attributes.
  pub fn get_attributes(&self) -> Result<WindowAttributes> {
    let attr = unsafe {
      libc::malloc(mem::size_of::<xlib::XWindowAttributes>())
        as *mut xlib::XWindowAttributes
    };
    let result = unsafe {
      xlib::XGetWindowAttributes(self.display.as_raw(), self.inner, attr)
    };
    match result {
      0 => Err(Error::GetAttributesError),
      _ => Ok(WindowAttributes {
        display: self.display,
        inner: attr,
      }),
    }
  }

  /// Get the raw window handle
  pub fn as_raw(&self) -> xlib::Window {
    self.inner
  }

  /// Gets the image from this window using XGetImage.
  pub fn get_image(&self) -> Result<Image> {
    let attr = self.get_attributes()?;
    Drawable::get_image(
      self,
      Rectangle::new(
        attr.get_x(),
        attr.get_y(),
        attr.get_width(),
        attr.get_height(),
      ),
    )
  }

  /// Change window property
  // TODO: make it more general
  pub fn change_property(&self, key: &Atom, val: &Atom) {
    use std::mem::transmute;
    let v = val.as_raw();
    unsafe {
      xlib::XChangeProperty(
        self.display.as_raw(),
        self.inner,
        key.as_raw(),
        xlib::XA_ATOM,
        32,
        xlib::PropModeReplace,
        transmute(&v),
        1,
      );
    }
  }

  /// "Maps", or shows a window to a display.
  pub fn map(&self) {
    unsafe { xlib::XMapWindow(self.display.as_raw(), self.inner) };
  }

  /// Requests for events from the server matching a particular event mask.
  ///
  /// If this function is not called, events will not be reported.
  pub fn select_input(&self, event_mask: EventMask) {
    debug!("event mask: {:?}", event_mask);
    unsafe {
      xlib::XSelectInput(
        self.display.as_raw(),
        self.inner,
        event_mask.bits() as i64,
      )
    };
  }
}

impl<'a> GetDisplay for Window<'a> {
  fn get_display(&self) -> *mut xlib::Display {
    self.display.as_raw()
  }
}

impl<'a> Drawable for Window<'a> {
  fn as_drawable(&self) -> xlib::Drawable {
    self.inner
  }
}

impl<'a> Drop for Window<'a> {
  fn drop(&mut self) {
    unsafe { xlib::XDestroyWindow(self.display.as_raw(), self.inner) };
  }
}

/// Window Attributes
pub struct WindowAttributes<'a> {
  pub(super) display: &'a Display,
  pub(self) inner: *mut xlib::XWindowAttributes,
}

impl<'a> WindowAttributes<'a> {
  /// Gets the width of the window
  pub fn get_x(&self) -> i32 {
    unsafe { (*self.inner).x as i32 }
  }

  /// Gets the height of the window
  pub fn get_y(&self) -> i32 {
    unsafe { (*self.inner).y as i32 }
  }

  /// Gets the width of the window
  pub fn get_width(&self) -> u32 {
    unsafe { (*self.inner).width as u32 }
  }

  /// Gets the height of the window
  pub fn get_height(&self) -> u32 {
    unsafe { (*self.inner).height as u32 }
  }

  /// Get the root window of this window
  pub fn get_root(&self) -> Window {
    Window {
      display: self.display,
      inner: unsafe { (*self.inner).root },
    }
  }
}

impl<'a> Drop for WindowAttributes<'a> {
  fn drop(&mut self) {
    unsafe { libc::free(self.inner as *mut libc::c_void) };
  }
}

bitflags! {
    #[derive(Default, Debug)]
    pub struct EventMask: c_ulong {
        const NO_EVENT_MASK = 0;
        const KEY_PRESS_MASK = (1<<0);
        const KEY_RELEASE_MASK = (1<<1);
        const BUTTON_PRESS_MASK = (1<<2);
        const BUTTON_RELEASE_MASK = (1<<3);
        const ENTER_WINDOW_MASK = (1<<4);
        const LEAVE_WINDOW_MASK = (1<<5);
        const POINTER_MOTION_MASK = (1<<6);
        const POINTER_MOTION_HINT_MASK = (1<<7);
        const BUTTON_1_MOTION_MASK = (1<<8);
        const BUTTON_2_MOTION_MASK = (1<<9);
        const BUTTON_3_MOTION_MASK = (1<<10);
        const BUTTON_4_MOTION_MASK = (1<<11);
        const BUTTON_5_MOTION_MASK = (1<<12);
        const BUTTON_MOTION_MASK = (1<<13);
        const KEYMAP_STATE_MASK = (1<<14);
        const EXPOSURE_MASK = (1<<15);
        const VISIBILITY_CHANGE_MASK = (1<<16);
        const STRUCTURE_NOTIFY_MASK = (1<<17);
        const RESIZE_REDIRECT_MASK = (1<<18);
        const SUBSTRUCTURE_NOTIFY_MASK = (1<<19);
        const SUBSTRUCTURE_REDIRECT_MASK = (1<<20);
        const FOCUS_CHANGE_MASK = (1<<21);
        const PROPERTY_CHANGE_MASK = (1<<22);
        const COLORMAP_CHANGE_MASK = (1<<23);
        const OWNER_GRAB_BUTTON_MASK = (1<<24);
    }
}

#[derive(Debug)]
pub struct SetWindowAttributes<'a> {
  pub background_pixmap: BackgroundPixmap<'a>,
  pub background_pixel: c_ulong,
  pub border_pixmap: Inherit<Pixmap<'a>>,
  pub border_pixel: c_ulong,
  pub bit_gravity: i32,
  pub win_gravity: i32,
  pub backing_store: BackingStore,
  pub backing_planes: u64,
  pub backing_pixel: u64,
  pub save_under: bool,
  pub event_mask: EventMask,
  pub do_not_propagate_mask: c_long,
  pub override_redirect: bool,
  pub colormap: Inherit<Colormap<'a>>,
  pub cursor: Noneable<Cursor<'a>>,
}

impl<'a> Default for SetWindowAttributes<'a> {
  fn default() -> Self {
    SetWindowAttributes {
      background_pixmap: BackgroundPixmap::default(),
      background_pixel: 0,
      border_pixmap: Inherit::default(),
      border_pixel: 0,
      bit_gravity: xlib::ForgetGravity,
      win_gravity: xlib::NorthWestGravity,
      backing_store: BackingStore::default(),
      backing_planes: 0xffffffff,
      backing_pixel: 0,
      save_under: false,
      event_mask: EventMask::empty(),
      do_not_propagate_mask: 0,
      override_redirect: false,
      colormap: Inherit::default(),
      cursor: Noneable::default(),
    }
  }
}

impl<'a> SetWindowAttributes<'a> {
  pub fn into_raw(self) -> *mut xlib::XSetWindowAttributes {
    let result = xlib::XSetWindowAttributes {
      background_pixmap: self.background_pixmap.into_int(),
      background_pixel: self.background_pixel,
      border_pixmap: self.border_pixmap.into_int(),
      border_pixel: self.border_pixel,
      bit_gravity: self.bit_gravity,
      win_gravity: self.win_gravity,
      backing_store: self.backing_store.into_raw(),
      backing_planes: self.backing_planes,
      backing_pixel: self.backing_pixel,
      save_under: self.save_under.into(),
      event_mask: self.event_mask.bits() as i64,
      do_not_propagate_mask: self.do_not_propagate_mask,
      override_redirect: self.override_redirect.into(),
      colormap: self.colormap.into_int(),
      cursor: self.cursor.into_int(),
    };

    Box::into_raw(Box::new(result))
  }
}

#[derive(Debug)]
pub enum BackgroundPixmap<'a> {
  Background(Pixmap<'a>),
  None,
  ParentRelative,
}

impl<'a> Default for BackgroundPixmap<'a> {
  fn default() -> Self {
    BackgroundPixmap::None
  }
}

impl<'a> IntoInt for BackgroundPixmap<'a> {
  fn into_int(&self) -> u64 {
    match self {
      BackgroundPixmap::Background(pixmap) => pixmap.inner,
      BackgroundPixmap::None => 0,
      BackgroundPixmap::ParentRelative => xlib::ParentRelative as u64,
    }
  }
}

#[derive(Debug)]
pub enum BackingStore {
  NotUseful,
  WhenMapped,
  Always,
}

impl Default for BackingStore {
  fn default() -> Self {
    BackingStore::NotUseful
  }
}

impl BackingStore {
  fn into_raw(self) -> i32 {
    match self {
      BackingStore::NotUseful => xlib::NotUseful,
      BackingStore::WhenMapped => xlib::WhenMapped,
      BackingStore::Always => xlib::Always,
    }
  }
}
